from conan.internal.model.conan_file import ConanFile
from conan.internal.model.workspace import Workspace
from conan.internal.model.version import Version

__version__ = '2.16.1'
conan_version = Version(__version__)
